/*
 mixbit.c
  main source file for the "mixbit" layered sound format

  Version 0.0.0.7

  Created: February 3, 2006
  Revised: May 15, 2006

 Copyright (C) 2006 by Adam Pickett

    This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

    This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.

    You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

May 5, 2006
A.P.: as of today, there will always be a constant 32 layers available
*/

#include <cseries.h>
#include <cs_system.h>
#include <cs_math.h>

#include "mixbit.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* constants */

/* version one bit width (positive and negative scales) */
#define MXB_ONE_WIDTH					15

#define MXB_RES_WIDTH					14
#define MXB_RES_SHIFT					(32-MXB_RES_WIDTH)

/* only applies to complex or animated waves */
#define MXB_AMP_SHIFT					12

static const char	nl[] = "negative layer?";

/* globals */

/* total size for all three tables is currently 80KB */
static SInt16		*mxb_triangle_table = (SInt16*)NULL;
static SInt16		*mxb_sine_table = (SInt16*)NULL;
/*static SInt16 *mxb_ramp_table = (SInt16*)NULL;*/

#include "random_tables.h"

/* macros */

#if TARGET_RT_LITTLE_ENDIAN
# define readU16(d,s,i)					do{d=(*s++<<8)+*s+++(i);}while(0)
#else
# define readU16(d,s,i)					do{d=(i)+*(UInt16*)s; s+=2;}while(0)
#endif
#if TARGET_RT_LITTLE_ENDIAN
# define readU24(d,s,w)					do{d=(*s++<<(16+(w)))|(*s++<<(8+(w)))|(*s++<<(w));}while(0)
#else
# define readU24(d,s,w)					do{d=((*(UInt32*)(-1+s))<<(w))&(0x00FFFFFF<<(w)); s+=3;}while(0)
#endif
#if TARGET_RT_LITTLE_ENDIAN
# define readU32(d,s,i)					do{d=(*s++<<24)|(*s++<<16)|(*s++<<8)|*s++; if((d+=i)==0) return -1;}while(0)
#else
# define readU32(d,s,i)					do{d=*(UInt32*)s; s+=4; if((i)==0) d+=i; else if((d+=i)==0) return -1;}while(0)
#endif

#if 1 /* 1 for (half-sample) linear interpolation */
# define TBLLOOKUP(t,x,s,r)				(((t)[(x)>>(s)]+(t)[((x)+(r))>>(s)])>>1)
#else
# define TBLLOOKUP(t,x,s,r)				((t)[(x)>>(s)])
#endif

#define TRILOOKUP(x)					TBLLOOKUP(mxb_triangle_table,x,MXB_RES_SHIFT,1L<<MXB_RES_SHIFT)
#define SINLOOKUP(x)					TBLLOOKUP(mxb_sine_table,x,MXB_RES_SHIFT,1L<<MXB_RES_SHIFT)
#define RAMPCALCULATE(x)				(SInt16)(((signed)x>=0)?((x)&((1L<<31)-1))>>(33-MXB_ONE_WIDTH):((SInt32)-((x)&((1L<<31)-1)))>>(33-MXB_ONE_WIDTH))

#define RNDLOOKUP(x)					(SInt16)((TBLLOOKUP(mxb_cliff_table,x,MXB_RES_SHIFT,1L<<MXB_RES_SHIFT)<<8)+mxb_cliff_table[((x)+0x55555555)>>MXB_RES_SHIFT])
#define WAVLOOKUP(x)					TBLLOOKUP(wave_table,x,wave_shift,temp)

#define MXB_GEN_FREQ_LOOP(x,y)		do{amp=1+*s++; readU32(fq,s,1); \
	do{if(d>=e) {tr1=tr2; break;} *d+++=(amp*(SInt16)(x(theta)))>>8; theta+=fq; y;}while(--count);}while(0)

#define MXB_GEN_ACCUM_LOOP(x,y,z)	do{sla=1+*s++; amp=(1+*s++)<<((z)-8); \
	acct=*s++<<13;readU32(fq,s,1); do{if(d>=e) {tr1=tr2; break;} \
	*d+++=(amp*(SInt16)(x(theta)))>>(z); theta+=fq; y; fq+=acct,amp+=sla;}while(--count);}while(0)

/* private prototypes */

static void build_mxb_tables(void);
static sfract fast_sine(ufract theta);

/* code */

long decode_mixbit_stream16(mixbit_header *header,
	void *inStream,
	void *outStream,
	UInt32 outLength,
	mixbit_layer_setup *layers)
{
	SInt16			*wave_table = NULL;

	const UInt8		*s, *tr1, *tr2 = NULL;
	UInt8			v, wave_shift;
	SInt16			*d, *_d, *e;
	UInt32			theta = 0, fq = 0, acct = 0, temp;
	int				count;
	short_int		mchan, cchan = 0;
	short			amp = 0x1000, sla, layer = 0;
	boolean			ovfl;

	if (mxb_triangle_table==(SInt16*)NULL)
		build_mxb_tables();
	if (mxb_triangle_table==(SInt16*)NULL)
		return kCS_memory_full_error; /* build_mxb_tables obviously failed */

	if ((s=(UInt8*)inStream)==(UInt8*)NULL || (s=(UInt8*)inStream)==(UInt8*)NONE
			|| layers==(mixbit_layer_setup*)NULL)
		return -1;
	if ((d=(SInt16*)outStream)==(SInt16*)NULL || (d=(SInt16*)outStream)==(SInt16*)NONE)
		return -1;
	if (outLength>0x7FFFFFF0 || (header->layer_info & 0xF8)!=(UInt8)0)
		return -1;
	if (header->version!=_mixbit_version_one)
		return kCS_unrecognized_version;

	/* initialize the layer offsets if the first position is negative */
	if (layers->postn[0]<0)
	{
		memset(layers->postn,0,sizeof(layers->postn));
		memset(layers->theta,0,sizeof(layers->theta));
		memset(layers->wavet,(int)NULL,sizeof(layers->wavet));
		memset(layers->waves,-1,sizeof(layers->waves));
	}

//	layer_extra = 0xF; layer_extra & header->layer_info;
	mchan = header->layer_info;
	e = (SInt16*)(outLength+(byte*)d);
	tr1 = s;
	while(1)
	{
		if (tr1!=tr2)
			{tr1 = tr2; tr2 = s;}
		while((v=*s++)<0xF0)
		{
			count = v & 0xF;
			if (v & 0x80)
				{count = (count<<8)+*s++; count = 16+(count<<8)+*s++;}
			++count;
			if (mchan<=0)
				switch(v & 0x70)
				{
				case 0x00:
					return -1;
				case 0x10:
					if (wave_table==(SInt16*)NULL)
						return -1;
					temp = 1L<<wave_shift-1;
					MXB_GEN_FREQ_LOOP(WAVLOOKUP,);
					break;
				case 0x20:
					MXB_GEN_FREQ_LOOP(RNDLOOKUP,);
					break;
				case 0x30: /* triangle */
					MXB_GEN_FREQ_LOOP(TRILOOKUP,);
					break;
				case 0x40: /* sine */
					MXB_GEN_FREQ_LOOP(SINLOOKUP,);
					break;
				case 0x50: /* ramp */
					MXB_GEN_FREQ_LOOP(RAMPCALCULATE,);
					break;
				case 0x60: /* animated sine */
					MXB_GEN_ACCUM_LOOP(SINLOOKUP,,MXB_AMP_SHIFT);
					break;
				case 0x70: /* sampled sound */
					_d = d;
					if ((s=mxb_decode_ssnd16(s,&_d,e,count,mchan,&ovfl))==NULL)
						return -1;
					d = _d;
					if (ovfl)
						tr1 = tr2;
				}
			else
				switch(v & 0x70)
				{
				case 0x00:
					return -1;
				case 0x10:
					if (wave_table==(SInt16*)NULL)
						return -1;
					temp = 1L<<wave_shift-1;
					MXB_GEN_FREQ_LOOP(WAVLOOKUP,d+=mchan);
					break;
				case 0x20:
					MXB_GEN_FREQ_LOOP(RNDLOOKUP,d+=mchan);
					break;
				case 0x30: /* triangle */
					MXB_GEN_FREQ_LOOP(TRILOOKUP,d+=mchan);
					break;
				case 0x40: /* sine */
					MXB_GEN_FREQ_LOOP(SINLOOKUP,d+=mchan);
					break;
				case 0x50: /* ramp */
					MXB_GEN_FREQ_LOOP(RAMPCALCULATE,d+=mchan);
					break;
				case 0x60: /* animated sine */
					MXB_GEN_ACCUM_LOOP(SINLOOKUP,d+=mchan,MXB_AMP_SHIFT);
					break;
				case 0x70: /* sampled sound */
					_d = d;
					if ((s=mxb_decode_ssnd16(s,&_d,e,count,mchan,&ovfl))==NULL)
						return -1;
					d = _d;
					if (ovfl)
						tr1 = tr2;
				}
		}
		switch(v)
		{
		case 0xF0: /* string */
			++s; /* ignored */
			s += *s++;
			break;
		case 0xF1: /* quadilateral */
			/* because the previous sample is used, it must exist */
			if ((byte*)d<=(byte*)outStream)
				return -1;
			temp = d[-1]<<16; /* sample */

			readU24(acct,s,8); acct = (signed)acct>>8;
			readU32(fq,s,0);
			readU32(count,s,0); ++count;
			do
			{
				if (d>=e)
					{tr1=tr2; break;}
				*d++ += (signed)(temp+=(fq+=acct))>>16;
			} while(--count);
			break;
		case 0xF2: /* set theta (UInt32)<<8 */
			readU32(theta,s,0);
			break;
		case 0xF3:
			if ((wave_shift=1+*s++)>23)
				return -1;
			wave_table = layers->wavet[layer] = (SInt16*)s;
			s += 2L<<wave_shift;
			layers->waves[layer] = wave_shift = 32-wave_shift;
			break;
		case 0xF4: /* previous layer */
			if (layer<0)
				{dprintf(nl); return -1;}
			layers->postn[layer] = (long)d-(long)outStream;
			layers->theta[layer] = theta;
			layer = --layer & 15;
			d = (mchan!=0 ? layer : 0)+(SInt16*)(layers->postn[layer]+((byte*)outStream));
			theta = layers->theta[layer];
			wave_table = layers->wavet[layer], wave_shift = layers->waves[layer];
			break;
		case 0xF5: /* next layer */
			if (layer<0)
				{dprintf(nl); return -1;}
			layers->postn[layer] = (long)d-(long)outStream;
			layers->theta[layer] = theta;
//			if (++layer>15)
//				return (long)d-(long)outStream;
			layer = ++layer & 15;
			d = (mchan!=0 ? layer : 0)+(SInt16*)(layers->postn[layer]+((byte*)outStream));
			theta = layers->theta[layer];
			wave_table = layers->wavet[layer], wave_shift = layers->waves[layer];
			break;
		case 0xF7: /* next channel, clear layers */
			memset(layers->postn,0,sizeof(layers->postn));
			if (++cchan>mchan)
				goto end;
			d = cchan+(SInt16*)outStream;
			break;
		case 0xF6: /* next channel */
			++d;
			if (++cchan>mchan)
				goto end;
			break;

		case 0xFF: /* terminate */
			goto end;
		}
	}
end:
	/* return the remaining data, if any */
	return (long)tr2-(long)inStream;
}

/* private code */

static void free_mxb_tables(void);

static void build_mxb_tables()
{
	ufract			theta, delta = 1L<<32-MXB_RES_WIDTH;
	sfract			x;

	SInt16			*t, *s;
	int				i;

	mxb_triangle_table = (SInt16*)realloc(mxb_triangle_table,2L<<MXB_RES_WIDTH);
	mxb_sine_table = (SInt16*)realloc(mxb_sine_table,2L<<MXB_RES_WIDTH);

	if (mxb_triangle_table==(SInt16*)NULL || mxb_sine_table==(SInt16*)NULL)
	{ /* failed, free everything */
		free_mxb_tables();
		return;
	}

	t = mxb_triangle_table, s = mxb_sine_table; theta = 0;
	i = 1L<<MXB_RES_WIDTH-1;
	do
	{
		/* calculate and round */
		*s++ = (fast_sine(theta)+(1L<<(30-15)))>>(30-14);
		*t++ = (0x40000000UL+(1L<<(30-15))-ABS(theta-0x40000000UL))>>(30-14);
		theta += delta;
	} while(--i);
	i = 1L<<MXB_RES_WIDTH-1;
	do
	{
		*s++ = (fast_sine(theta)+(1L<<(30-15))-1)>>(30-14);
		*t++ = (0xC0000000UL+(1L<<(30-15))-1+ABS(theta-0xC0000000UL))>>(30-14);
		theta += delta;
	} while(--i);
}

static void free_mxb_tables()
{
	if (mxb_triangle_table!=(SInt16*)NULL)
		free((void*)mxb_triangle_table);
	if (mxb_sine_table!=(SInt16*)NULL)
		free((void*)mxb_sine_table);
/*	if (mxb_ramp_table!=(SInt16*)NULL)
		free((void*)mxb_ramp_table);*/
}

/* fract trig routines */

/* this LUT is a list of the fraction other than theta in the power series
  x**3/6 - ...
*/
static const UInt32 _ufract_sine_table[257] =
{
	0xA55DE731, 0xA55DD2CA, 0xA55D9596, 0xA55D2F95,
	0xA55CA0C7, 0xA55BE92C, 0xA55B08C5, 0xA559FF92,
	0xA558CD95, 0xA55772CE, 0xA555EF3E, 0xA55442E7,
	0xA5526DC8, 0xA5506FE5, 0xA54E493D, 0xA54BF9D3,
	0xA54981A8, 0xA546E0BD, 0xA5441716, 0xA54124B4,
	0xA53E0998, 0xA53AC5C5, 0xA537593D, 0xA533C404,
	0xA530061A, 0xA52C1F83, 0xA5281042, 0xA523D85A,
	0xA51F77CD, 0xA51AEE9E, 0xA5163CD1, 0xA5116269,
	0xA50C5F6A, 0xA50733D7, 0xA501DFB3, 0xA4FC6303,
	0xA4F6BDCA, 0xA4F0F00C, 0xA4EAF9CD, 0xA4E4DB12,
	0xA4DE93DF, 0xA4D82438, 0xA4D18C22, 0xA4CACBA1,
	0xA4C3E2BA, 0xA4BCD173, 0xA4B597CF, 0xA4AE35D5,
	0xA4A6AB88, 0xA49EF8F0, 0xA4971E10, 0xA48F1AEF,
	0xA486EF92, 0xA47E9BFF, 0xA476203C, 0xA46D7C4E,
	0xA464B03C, 0xA45BBC0D, 0xA4529FC5, 0xA4495B6D,
	0xA43FEF09, 0xA4365AA2, 0xA42C9E3C, 0xA422B9E1,
	0xA418AD96, 0xA40E7962, 0xA4041D4C, 0xA3F9995D,
	0xA3EEED9B, 0xA3E41A0D, 0xA3D91EBC, 0xA3CDFBAF,
	0xA3C2B0EE, 0xA3B73E81, 0xA3ABA470, 0xA39FE2C2,
	0xA393F981, 0xA387E8B4, 0xA37BB065, 0xA36F509B,
	0xA362C95F, 0xA3561ABB, 0xA34944B6, 0xA33C475B,
	0xA32F22B1, 0xA321D6C3, 0xA314639A, 0xA306C93E,
	0xA2F907BB, 0xA2EB1F18, 0xA2DD0F60, 0xA2CED89E,
	0xA2C07AD9, 0xA2B1F61E, 0xA2A34A76, 0xA29477EB,
	0xA2857E88, 0xA2765E56, 0xA2671762, 0xA257A9B4,
	0xA2481559, 0xA2385A5B, 0xA22878C5, 0xA21870A2,
	0xA20841FE, 0xA1F7ECE3, 0xA1E7715D, 0xA1D6CF78,
	0xA1C6073F, 0xA1B518BE, 0xA1A40401, 0xA192C913,
	0xA1816801, 0xA16FE0D7, 0xA15E33A1, 0xA14C606B,
	0xA13A6742, 0xA1284832, 0xA1160348, 0xA1039891,
	0xA0F1081A, 0xA0DE51EF, 0xA0CB761D, 0xA0B874B2,
	0xA0A54DBB, 0xA0920145, 0xA07E8F5E, 0xA06AF813,
	0xA0573B71, 0xA0435987, 0xA02F5263, 0xA01B2611,
	0xA006D4A1, 0x9FF25E21, 0x9FDDC29E, 0x9FC90226,
	0x9FB41CC9, 0x9F9F1295, 0x9F89E399, 0x9F748FE2,
	0x9F5F1781, 0x9F497A83, 0x9F33B8F8, 0x9F1DD2EF,
	0x9F07C877, 0x9EF199A0, 0x9EDB4678, 0x9EC4CF0F,
	0x9EAE3375, 0x9E9773BA, 0x9E808FEC, 0x9E69881C,
	0x9E525C5B, 0x9E3B0CB7, 0x9E239941, 0x9E0C020A,
	0x9DF44721, 0x9DDC6897, 0x9DC4667C, 0x9DAC40E2,
	0x9D93F7D8, 0x9D7B8B6F, 0x9D62FBB9, 0x9D4A48C6,
	0x9D3172A7, 0x9D18796E, 0x9CFF5D2B, 0x9CE61DF0,
	0x9CCCBBCE, 0x9CB336D7, 0x9C998F1C, 0x9C7FC4B0,
	0x9C65D7A2, 0x9C4BC807, 0x9C3195EF, 0x9C17416C,
	0x9BFCCA91, 0x9BE2316F, 0x9BC7761A, 0x9BAC98A3,
	0x9B91991C, 0x9B767799, 0x9B5B342C, 0x9B3FCEE7,
	0x9B2447DE, 0x9B089F22, 0x9AECD4C8, 0x9AD0E8E3,
	0x9AB4DB84, 0x9A98ACC0, 0x9A7C5CAA, 0x9A5FEB55,
	0x9A4358D4, 0x9A26A53C, 0x9A09D09F, 0x99ECDB12,
	0x99CFC4A8, 0x99B28D75, 0x9995358D, 0x9977BD03,
	0x995A23ED, 0x993C6A5E, 0x991E906B, 0x99009627,
	0x98E27BA8, 0x98C44101, 0x98A5E648, 0x98876B90,
	0x9868D0EF, 0x984A167A, 0x982B3C44, 0x980C4264,
	0x97ED28EF, 0x97CDEFF8, 0x97AE9796, 0x978F1FDE,
	0x976F88E5, 0x974FD2C1, 0x972FFD87, 0x9710094C,
	0x96EFF626, 0x96CFC42B, 0x96AF7371, 0x968F040E,
	0x966E7616, 0x964DC9A2, 0x962CFEC5, 0x960C1597,
	0x95EB0E2E, 0x95C9E8A0, 0x95A8A503, 0x9587436E,
	0x9565C3F8, 0x954426B6, 0x95226BC0, 0x9500932B,
	0x94DE9D10, 0x94BC8984, 0x949A589F, 0x94780A78,
	0x94559F25, 0x943316BE, 0x94107159, 0x93EDAF0F,
	0x93CACFF6, 0x93A7D425, 0x9384BBB5, 0x936186BC,
	0x933E3552, 0x931AC790, 0x92F73D8B, 0x92D3975D,
	0x92AFD51C, 0x928BF6E2, 0x9267FCC5, 0x9243E6DE,
	0x921FB544
};

/*
 all values of [0,4.0) represent [0,2*pi)

   the larger theta is, the less accurate the result (in between the entries
 in the table), although it is still accurate enough for 16 bit samples
*/
static sfract fast_sine(ufract theta)
{
	ufract			sa, sb;
	unsigned int	t, u;

	/*
	theta' = theta{30}

	high two bits:
		00 (0.0pi): theta'' =  theta';    result =  sin(theta'')
		01 (0.5pi): theta'' =  pi-theta'; result =  sin(theta'')
		10 (1.0pi): theta'' =  theta';    result = -sin(theta'')
		11 (1.5pi): theta'' =  pi-theta'; result = -sin(theta'')
	*/
	t = (1L<<30)-1; t &= theta;
	if (theta & (1L<<30))
		{t = -t; t += (1L<<30);}

	u = t>>22;
	/* get the two nearest values in the LUT of (theta mod pi/2) and interpolate */
	sb = _ufract_sine_table[u];
	sa = sb-_ufract_sine_table[u+1];
	sb -= UFract32Mul(sa,t<<10);

	/* theta''*(pi/2-theta''**2*...) */
	sa = UFract32Mul(t,t)<<3;
	sa = 0xC90FDAA2-UFract32Mul(sa,sb), t <<= 1;
	sa = UFract32Mul(sa,t);

	/* negative if >= pi */
	if (theta & (1L<<31))
		return -sa;
	else
		return sa;
}
