/*
 mixbit.h
  source code header for my new "mixbit" encoding

  Version 0.0.0.9

  Created: February 3, 2006
  Revised: June 1, 2006
*/

/* another faux-franchise? */

/*
  there are a maximum of 32 layers; crossing the border will wrap

  the low four bits in the entry header are primarily the bits for the entry length
    (plus 32).

  if bit #7 is one, then an extra 16 bits adds to the sample count width and 16
    is also added to the count; making a total range of [32,1048623).

   0x0x - opcode:
     0x00, terminator
     0x01, quadrilateral    -- accslop. S24, S32, sample count U32+1
     0x02, set theta        -- U32
     0x03, set wave table   -- bit width (U8+1), table (S16); extra data is ignored
     0x04, previous layer   -- wraps around
     0x05, next layer       -- wraps around
     0x06, next channel     -- terminate on wrap; continues w/current layers
     0x07, next channel+clr -- terminate on wrap; clears all layer offsets
     0x08-0x0E, reserved
     0x0F, Pascal string    -- type U8, string; -- doesn't directly affect the sound
   0x1x - wave table        -- amp. U8+1; frq. (U32+1);
   0x2x - cliff noise       -- amp. U8+1; frq. (U32+1);
   0x3x - triangle          -- amp. U8+1; frq. (U32+1);
   0x4x - sine              -- amp. U8+1; frq. (U32+1);
   0x5x - ramp              -- amp. U8+1; frq. (U32+1);
   0x6x - animated sine     -- slop/amp. U8+1, (U8+1)<<4; acc/frq. U8+1<<13, (U32+1)
   0x7x - sampled sound     -- format (U8); (if bit 7 set) [U24+1 byte count]

  if the outgoing buffer has a limit then overflow forces the decoder to switch to
    scanning past all generationg opcodes until it has reached a layer/channel
    that hasn't yet overflowed

  the "quadrilateral" opcode generates a curve using the previous sample -- it may
    wrap on sample boundary; this generation opcode would also be used to encode
    silence in layers
*/

/* constants */

#define _mixbit_version_one         0x15 /* 15 bit */

#if !defined(rez) && !defined(derez)

/* string opcode types */
#define _mixbit_string_comment      0x00
#define _mixbit_string_name         0x01
#define _mixbit_string_author       0x02
#define _mixbit_string_copyright    0x03

#define _mixbit_string_user_data    0x80

#define _mixbit_string_CRC8         0xFF
#define _mixbit_string_CRC16        0xFE
#define _mixbit_string_CRC24        0xFD
#define _mixbit_string_CRC32        0xFC
#define _mixbit_string_CRC64        0xFB

/* the structures are all big-endian */

/* layer_info = [invalid.5] [channel_count.3]+1 -- max. of 128 layers */
typedef struct mixbit_format_header /* (12) */
{
   UInt8            version, layer_info;
   UInt16           sample_rate; /* +1000 */
   UInt64           projected_length; /* total sample count */
} mixbit_header;

typedef struct mixbit_seek_sample /* (16) */
{
   UInt64           sample_level; /* absolute sample (data format applies) */
   UInt64           sample_offset; /* absolute offset in file */
} mixbit_seek;

typedef struct mixbit_layer_setup
{
   long          postn[32];
   UInt32        theta[32];
   UInt8         waves[32];
   SInt16        *wavet[32];
} mixbit_layer_setup;

/* prototypes */

/*
 decode_mixbit_stream

  This routine ignores the sample_rate and length fields.
  There's no limit on `inStream', so make sure it's terminated with 0xFF.

  A non-positive result is an error code, a positive result is the offset within
    the `inStream' where it was terminated if overflow did not occur -- if
    overflow occurs, the result is the last position before overflow, so that
    only part of the stream is needed to be repeated.

  IMPORTANT: It adds to the already existing sample data, so if you are filling
    from nothing then it is best if you clear the data before calling.
*/

/* 16 bit signed */
long decode_mixbit_stream16(mixbit_header *header,
  void *inStream,
  void *outStream,
  UInt32 outLength,
  mixbit_layer_setup *layers);

/* decodes a sampled sound chunk; returns the next offset or NULL if invalid;
  mixes to pointer in `d' */
const UInt8 *mxb_decode_ssnd16(const UInt8 *s,
  SInt16 **d,
  SInt16 *e,
  int c,
  short depth,
  boolean *ovfl); /* can be NULL */

/* encodes sampled sound of `format' (read mixbit_ssnd.h); returns the end point
  set bit #8 for `format' to mirror overflow (to avoid clipping) */
UInt8 *mxb_encode_samples16(const SInt16 *_s,
  UInt8 *d,
  int count,
  int format,
  short depth); /* zero for no depth (mono) */
UInt8 *mxb_encode_samples8u(const UInt8 *_s,
  UInt8 *d,
  int count,
  int format,
  short depth); /* zero for no depth (mono) */

/* returns the maximum (fractional) byte size per sample;
  returns zero if the passed value the format is not available */
ufixed get_ssnd_maximum_size(int format);

#else /* !defined(rez) && !defined(derez) */

# define _sCc mono = 0, stereo = 1

# define _sRs rate8k = 7000, rate11025 = 10025, rate15k = 14000, rate16k = 15000, \
   rate22050 = 21050, rate22k = 21255, rate24k = 23000, rate27900 = 26900, \
   rate32k = 31000, rate44k = 43100, rate46200 = 45200, rate48k = 47000, \
   rate56k = 55000, rate60k = 63000, rate64k = 63000, rate66k = 65000

  type 'mxsn'
  {
    switch   /* version */
    {
    case VersionOne:
      key byte = _mixbit_version_one;
      unsigned bitstring[5] = 0;  /* validity check */
      unsigned bitstring[3] _sCc; /* channel count (+1) */
      unsigned integer _sRs;      /* sample rate (+1000) */

      unsigned longint = 0;
      unsigned longint;           /* length -- lower 32 bits, of course */

      hexstring;    /* data */
    };
  };

#undef _sCc
#undef _sLc
#undef _sRs

#endif
