/*
	IAScript Execution Header

	For version 0.02 (not completely implemented)

	Revised: July 15, 2005

NOTE: Your program source code does not need to include this file.
	This is meant for the scripting internal code only.
*/

enum
{
/* identifier flags - SLOT_USED also applies */
	_sid_is_built_in			= 0x0001,			/* came from initial_scripting_ids.h */
	_sid_is_read_at_reset		= 0x0002,			/* changing it is meaningless until reset */
	_sid_is_written_to_disk		= 0x0004,			/* gets written to file on program termination */
	_sid_is_boolean				= 0x0008,			/* zero is seen as false, and non-zero is seen as true */
	_sid_is_constant			= 0x0080			/* obviously not a variable */
};

/*
	each opcode consists of flags, opcode size, and an 8 bit value:
		XLCCZZZZ VVVVVVVV ... (parameters)

	X = external
	L = line
	C = condition bits
	Z = size
	V = value

	when X is one, V becomes three bytes in size instead of one byte
*/

enum
{
/* opcode bits */
	_op_size_mask				= 0x0F,		/* size of opcode is always 2+size*2 */
	_op_line_bit				= 0x8000,	/* the opcode was coded at the beginning of the line in the script */
	_op_condition_bit1			= 0x2000,	/* custom to the opcode */
	_op_condition_bit2			= 0x1000,	/* custom to the opcode */

	_op_external_bit			= 0x0080,	/* otherwise an internal function/operation */
	_op_return_opcode_mask		= 0x007F,	/* opcode for the delayed operation that came before it */
/*	_op_external_mask = 0x0000FFFF, */

/* parameter bits */
	_opp_identifier				= 0x8000,	/* high bit is a flag for whether it is an id or a literal */
	_opp_global					= 0x4000,	/* high bit is a flag for whether it points to a global id, otherwise local */
	_opp_index_mask				= 0x0FFF,	/* up to 4095 local ids */

/*
	 Except for the boolean type, literal parameters first carry a type byte and
	a size byte.  The type byte holds the same concept as the ones used in the
	identifier lists, but are custom to the size.  If a size is odd or is not
	applicable for the type, an error will result.
*/
	_opl_byte					= 0x0000,	/* low byte is the byte (unsigned) */
	_opl_short					= 0x0102,	/* followed by a two-byte literal (signed) */
	_opl_integer				= 0x0104,	/* followed by a four-byte literal (signed) */
	_opl_unsigned_short			= 0x0202,	/* followed by a two-byte literal (unsigned) */
	_opl_unsigned_integer	= 0x0204,	/* followed by a four-byte literal (unsigned) */
	_opl_unsigned_wide		= 0x0208,	/* followed by an eight-byte literal (unsigned) */
	_opl_single				= 0x0304,	/* followed by a 32-bit single (floating-point) */
	_opl_double				= 0x0308,	/* followed by a 64-bit single (floating-point) */
	_opl_string				= 0x0400,	/* c-string (up to 28 bytes in size) */

/* parameter codes */
	_opp_stack				= 0			/* pulls or pushes ids off the stack - applies any parameter */
};

/*
	  Parsed expressions usually come out popping a value off of the stack and
	and placing the result back on or in the destination register.
	  For example, for the expression:

		x = y+z

	if x, y, and z are local indices #1, #2, and #3, it would come out...

		8103 8001				dest		x
		0110 8002				push		y
		2120 8003				add		D	z	; pops y off the stack

	the alternative method will put it in a different order

		y+z->x

	will instead generate...

		8110 8002				push		y
		0120 8003				add			z	; replaces <y> on the stack with the sum
		2111 8001				pop			x

  ----

  Expresion Runs

  	Expression runs are when a certain operation is repeated in a consecutive fashion:

	 a = b+c+d+e+f+g+h+i+j+k+4

  if a-k are local indices #1-#11, the parser will generate...

	 8103 8001				dest		a
	 0110 8002				push		b
	 2A20 8003 8004 8005	add			c,d,e,f,g,h,i,j,k,4
	\ 8006 8007 8008 8009
	\ 800A 800B 0004

  ----

  Calling and Returning

	 result_code = 1+do_this(x,y,z,2);

	 8103 8001				dest		result_code		; sets DR to result_code
	 2118 0001				ls		D	#1				; sets <dest> to (1)
	 0105 2020				lop		D	add				; links the add opcode
	 0110 8002				push		x				; pushes x onto the stack
	 0110 8003				push		y				; pushes y onto the stack
	 0110 8004				push		z				; pushes z onto the stack
	 0118 0002				ls			#2				; pushes (2) onto the stack
	 0180 C040				call		do_this			; calls do_this
	 0003					dest						; clears DR

	  If the current link entry on the link stack is blank (has no return address),
  	then a call will set the address rather of the current one, rather than
  	pushing another onto the stack.
	  Function calls push the destination and the offset/line on the link stack,
	and a return pulls these values off and assigns the value on the stack
	to the id in the destination register (it stays on the stack if the
	destination register is zero).
*/

typedef enum
{
/*
	 When it comes to instructions that use <dest>, if the first condition bit is one
	then the stack is the destination, otherwise the destination register contains it.
	 Most instructions pull the first value off the stack.
*/
	_opc_return_to_process,	/* leaves execution (returning a zero line index) */
	_opc_return,			/* leaves function with (ids) -- nothing is taken as zero/empty */
	_opc_nop,				/* no operation; commented/blank lines are parsed as nops to increment the counter */
	_opc_dest,				/* sets the destination id to (id) */

	_opc_illegal,			/* returns the _srerr_illegal error */
	_opc_lop,				/* pushes a blank opcode onto the link stack */

/* 8 - id definition */
	_opc_defv		= 8,	/* creates local variable of (type) and new (id) -- error if the index already exists */
	_opc_defc,				/* creates local constant of (type) and new (id) -- error if the index already exists */
	_opc_undef,				/* removes local identifiers (returns an error if the global bit is set on any of the parameters) */
	_opc_redef,				/* converts a variable into a different type */

/* 16 - copy */
	_opc_move		= 16,	/* <dest> is set to value of (id) - (translates if theyre not the same type) */
	_opc_moved,				/* <dest> is set to data of (id) */
	_opc_moveb,				/* <dest> is set to byte index (id) of (id) */
	_opc_exg,				/* data of <dest> and (id) are exchanged */

	_opc_concat,			/* <dest> is set to (id) concatenated with (id) */

/* 24 - load (was set) */
	_opc_ls			= 24,	/* sets <data> to a literal (16 bit integer) */
	_opc_li,				/* ... (32 bit integer) */
	_opc_ll,				/* ... (64 bit integer) */
	_opc_ld,				/* ... (64 bit double) */

/* 32 - arithmetic */
	_opc_add		= 32,	/* sets <dest> to the sum of the others */
	_opc_sub,				/* sets <dest> to the difference of the others */
	_opc_neg,				/* sets <dest> to (id) subtracted from zero */
	_opc_abs,				/* sets <dest> to the absolute value of (id) */

	_opc_mul,				/* sets <dest> to product of the others */
	_opc_div,				/* sets <dest> to the division of the others */
	_opc_mod,				/* sets <dest> to the remainder of (id) divided by (id) */
	_opc_pow,				/* sets <dest> to (id) raised to the (id) power */
/* 40 - all values are truncated for logical and bitwise instructions */
	_opc_shl,				/* sets <dest> to (id) logical shift left of (id) bits */
	_opc_shr,				/* sets <dest> to (id) logical shift right of (id) bits */
	_opc_rol,				/* sets <dest> to (id) logical rotate left of (id) bits */
	_opc_ror,				/* sets <dest> to (id) logical rotate right of (id) bits */

	_opc_not,				/* sets <dest> to the bitwise NOT of (id) -- error if more than one parameter */
	_opc_or,				/* sets <dest> to the bitwise OR of the other parameters */
	_opc_and,				/* sets <dest> to the bitwise AND of the other parameters */
	_opc_xor,				/* sets <dest> to the bitwise EX-OR of the other parameters */
/* 48  - must have the minimum number of parameters */
	_opc_sbb,				/* set <dest> to true/one if (id) is either true/false, false/zero otherwise */
	_opc_sbr,				/* ... is neither true/false .... */
	_opc_sbf,				/* set <dest> to true/one if (id) is false, false/zero otherwise */
	_opc_sbt,				/* ... true/one if (id) is true ... */

	_opc_sbnan,				/* ... not a number ... */
	_opc_sbreal,			/* ... is a number ... */
	_opc_sbinf,				/* ... is infinite ... */
	_opc_sbninf,			/* ... is not infinite ... */
/* 56 */
	_opc_sbeq,				/* set <dest> to true/one if (id) is eqal to (id), false otherwise */
	_opc_sbne,				/* ... not equal ... */
	_opc_sbgl,				/* set <dest> to true/one if (id) cannot be compared to (id), false otherwise */
	_opc_sbnn,				/* ... can be compared to ... */

	_opc_sblt,				/* set <dest> to true/one if (id) is less than (id), false otherwise */
	_opc_sbgt,				/* ... greater than ... */
	_opc_sble,				/* ... less than or equal to ... */
	_opc_sbge,				/* ... greater than or equal to ... */
/* 64 */
/* boolean instructions use interpretation to whether the parameters are non-zero or true */
	_opc_bor,				/* sets <dest> to the boolean OR of the other parameters */
	_opc_band,				/* sets <dest> to the boolean AND of the other parameters */
	_opc_bnor,				/* sets <dest> to the boolean NOT-OR of the other parameters */
	_opc_bnand,				/* sets <dest> to the boolean NOT-AND of the other parameters */

	_opc_ext,				/* sets the current entry on the stack to sign extension, given size <condition bits> */
	_opc_count,				/* <dest> is set to the cstring length from the beginning of the data (text length/first zero byte) */
/*
	 Use of unimplemented opcodes can generate undefined results if not the
	unimplemented error.  But since text is the only way to input code by the
	non-programmer, this shouldnt be (much) of a problem.
*/
	_opc_push = _opc_move | 0x10000,				/* if one parameter -- error if no parameters */
	_opc_pop = _opc_moved | 0x10000,				/* if one parameter -- error if no parameters */

/* the jumping instructions use the low 32 bits of the ids value and return an error
	if the value is not an integer or the value is out of range
*/
	_opc_jump = _op_external_bit,					/* jump unconditionally to offset (id - low 32 bits of data) */
	_opc_jumpc = _opc_jump | _op_condition_bit2,	/* jump to offset on condition (id - non-zero) */
/* the calling instructions will return an error if the passed id is not a function */
	_opc_call = _opc_jump | _op_condition_bit1,		/* call unconditionally to function (id) */
	_opc_callc = _opc_call | _op_condition_bit2		/* call to offset on condition (id - non-zero) */
} internal_scripting_opcode;

/* structures */

typedef struct global_identifier//(32)
{
	UInt16			flags, type;
	byte			data[8];						/* may point to extra data */
	void			*callback;						/* gets called when the ID is used/changed */
	char			name[16];	//20-sizeof(void*)];
} global_identifier, global_id;

typedef struct link_entry//(16)
{
	/*
		destination register, return operation, stack pointer, local identifier pointer
		 - real data stack is data_stack+sp*sizeof(dual_value), real local id pointer
		   is scripting_temporary+lid*sizeof(dual_value)
		low order bits 7-15 of rop currently must be zero
		low order bits 5 and 6 of rop must not be zero
	*/
	UInt16			dr, rop, sp, lip;

	func_line		address;
} link_entry;

typedef struct local_identifier//(12)
{
	UInt16			flags, type;
	byte			data[8];
} local_identifier, local_id;
